Class: DEBUGGER__::ThreadClient
- Includes:
- Color, SkipPathHelper
- Defined in:
- lib/debug/server_cdp.rb,
lib/debug/server_dap.rb,
lib/debug/thread_client.rb
Defined Under Namespace
Classes: Recorder, SuspendReplay
Constant Summary collapse
- SPECIAL_LOCAL_VARS =
[ [:raised_exception, "_raised"], [:return_value, "_return"], ]
- SKIP_GLOBAL_LIST =
%i[$= $KCODE $-K $SAFE].freeze
Instance Attribute Summary collapse
-
#check_bp_fulfillment_map ⇒ Object
readonly
Returns the value of attribute check_bp_fulfillment_map.
-
#id ⇒ Object
readonly
Returns the value of attribute id.
-
#recorder ⇒ Object
readonly
Returns the value of attribute recorder.
-
#thread ⇒ Object
readonly
Returns the value of attribute thread.
Class Method Summary collapse
Instance Method Summary collapse
- #<<(req) ⇒ Object
- #assemble_arguments(args) ⇒ Object
- #class_method_map(classes) ⇒ Object
- #close ⇒ Object
- #collect_locals(frame) ⇒ Object
-
#constant_name?(name) ⇒ Boolean
TODO: support non-ASCII Constant name.
- #current_frame ⇒ Object
- #deactivate ⇒ Object
- #debug_cmd(cmds) ⇒ Object
- #debug_event(ev, args) ⇒ Object
- #debug_mode(old_mode, new_mode) ⇒ Object
- #debug_suspend(event) ⇒ Object
- #default_frame_formatter(frame) ⇒ Object
- #evaluate_result(r) ⇒ Object
- #event!(ev, *args) ⇒ Object
- #frame_eval(src, re_raise: false) ⇒ Object
- #frame_eval_core(src, b) ⇒ Object
- #frame_str(i, frame: ) ⇒ Object
- #generate_info ⇒ Object
- #get_frame(index) ⇒ Object
- #get_src(frame, max_lines:, start_line: nil, end_line: nil, dir: +1) ⇒ Object
-
#initialize(id, q_evt, q_cmd, thr = Thread.current) ⇒ ThreadClient
constructor
A new instance of ThreadClient.
- #inspect ⇒ Object
- #internalProperty(name, obj) ⇒ Object
- #location ⇒ Object
- #make_breakpoint(args) ⇒ Object
- #management? ⇒ Boolean
- #mark_as_management ⇒ Object
- #name ⇒ Object
- #on_breakpoint(tp, bp) ⇒ Object
- #on_init(name) ⇒ Object
- #on_load(iseq, eval_src) ⇒ Object
- #on_pause ⇒ Object
- #on_trace(trace_id, msg) ⇒ Object
- #on_trap(sig) ⇒ Object
- #outline_method(o, klass, obj) ⇒ Object
- #preview(name, obj) ⇒ Object
- #preview_(value, hash, overflow) ⇒ Object
- #process_cdp(args) ⇒ Object
- #process_dap(args) ⇒ Object
- #propertyDescriptor(name, obj) ⇒ Object
- #propertyDescriptor_(name, obj, type, description: nil, subtype: nil) ⇒ Object
- #puts(str = '') ⇒ Object
- #puts_variable_info(label, obj, pat) ⇒ Object
- #replay_suspend ⇒ Object
- #running? ⇒ Boolean
- #search_const(b, expr) ⇒ Object
- #set_mode(mode) ⇒ Object
-
#show_by_editor(path = nil) ⇒ Object
cmd: show edit.
- #show_consts(pat, only_self: false) ⇒ Object
- #show_frame(i = 0) ⇒ Object
-
#show_frames(max = nil, pattern = nil) ⇒ Object
cmd: show frames.
- #show_globals(pat) ⇒ Object
- #show_ivars(pat) ⇒ Object
- #show_locals(pat) ⇒ Object
-
#show_outline(expr) ⇒ Object
cmd: show outline.
- #show_src(frame_index: @current_frame_index, update_line: false, max_lines: , **options) ⇒ Object
-
#special_local_variables(frame) ⇒ Object
cmd: show.
- #step_tp(iter, events = [:line, :b_return, :return]) ⇒ Object
- #suspend(event, tp = nil, bp: nil, sig: nil, postmortem_frames: nil, replay_frames: nil, postmortem_exc: nil) ⇒ Object
- #to_s ⇒ Object
- #tp_allow_reentry ⇒ Object
- #truncate(string, width:) ⇒ Object
- #value_inspect(obj) ⇒ Object
- #variable(name, obj) ⇒ Object
- #variable_(name, obj, indexedVariables: 0, namedVariables: 0) ⇒ Object
- #wait_next_action ⇒ Object
- #wait_next_action_ ⇒ Object
-
#wait_reply(event_arg) ⇒ Object
events.
- #waiting? ⇒ Boolean
Methods included from SkipPathHelper
#skip_config_skip_path?, #skip_internal_path?, #skip_location?, #skip_path?
Methods included from Color
#color_pp, #colored_inspect, #colorize, #colorize_blue, #colorize_code, #colorize_cyan, #colorize_dim, #colorize_magenta, #irb_colorize, #with_inspection_error_guard
Constructor Details
#initialize(id, q_evt, q_cmd, thr = Thread.current) ⇒ ThreadClient
Returns a new instance of ThreadClient.
103 104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 |
# File 'lib/debug/thread_client.rb', line 103 def initialize id, q_evt, q_cmd, thr = Thread.current @is_management = false @id = id @thread = thr @target_frames = nil @q_evt = q_evt @q_cmd = q_cmd @step_tp = nil @output = [] @frame_formatter = method(:default_frame_formatter) @var_map = {} # { thread_local_var_id => obj } for DAP @obj_map = {} # { object_id => obj } for CDP @recorder = nil @mode = :waiting @current_frame_index = 0 # every thread should maintain its own CheckBreakpoint fulfillment state @check_bp_fulfillment_map = {} # { check_bp => boolean } set_mode :running thr.instance_variable_set(:@__thread_client_id, id) ::DEBUGGER__.info("Thread \##{@id} is created.") end |
Instance Attribute Details
#check_bp_fulfillment_map ⇒ Object (readonly)
Returns the value of attribute check_bp_fulfillment_map.
54 55 56 |
# File 'lib/debug/thread_client.rb', line 54 def check_bp_fulfillment_map @check_bp_fulfillment_map end |
#id ⇒ Object (readonly)
Returns the value of attribute id.
54 55 56 |
# File 'lib/debug/thread_client.rb', line 54 def id @id end |
#recorder ⇒ Object (readonly)
Returns the value of attribute recorder.
54 55 56 |
# File 'lib/debug/thread_client.rb', line 54 def recorder @recorder end |
#thread ⇒ Object (readonly)
Returns the value of attribute thread.
54 55 56 |
# File 'lib/debug/thread_client.rb', line 54 def thread @thread end |
Class Method Details
.current ⇒ Object
42 43 44 45 46 47 48 49 |
# File 'lib/debug/thread_client.rb', line 42 def self.current if thc = Thread.current[:DEBUGGER__ThreadClient] thc else thc = SESSION.get_thread_client Thread.current[:DEBUGGER__ThreadClient] = thc end end |
Instance Method Details
#<<(req) ⇒ Object
202 203 204 205 |
# File 'lib/debug/thread_client.rb', line 202 def << req debug_cmd(req) @q_cmd << req end |
#assemble_arguments(args) ⇒ Object
60 61 62 63 64 |
# File 'lib/debug/thread_client.rb', line 60 def assemble_arguments(args) args.map do |arg| "#{colorize_cyan(arg[:name])}=#{arg[:value]}" end.join(", ") end |
#class_method_map(classes) ⇒ Object
728 729 730 731 732 733 734 735 736 |
# File 'lib/debug/thread_client.rb', line 728 def class_method_map(classes) dumped = Array.new classes.reject { |mod| mod >= Object }.map do |mod| methods = mod.public_instance_methods(false).select do |m| dumped.push(m) unless dumped.include?(m) end [mod, methods] end.reverse end |
#close ⇒ Object
170 171 172 |
# File 'lib/debug/thread_client.rb', line 170 def close @q_cmd.close end |
#collect_locals(frame) ⇒ Object
514 515 516 517 518 519 520 521 522 523 524 525 526 527 528 529 530 531 |
# File 'lib/debug/thread_client.rb', line 514 def collect_locals(frame) locals = [] if s = frame&.self locals << ["%self", s] end special_local_variables frame do |name, val| locals << [name, val] end if vars = frame&.local_variables vars.each{|var, val| locals << [var, val] } end locals end |
#constant_name?(name) ⇒ Boolean
TODO: support non-ASCII Constant name
741 742 743 744 745 746 747 748 749 750 751 752 753 754 755 |
# File 'lib/debug/thread_client.rb', line 741 def constant_name? name case name when /\A::\b/ constant_name? $~.post_match when /\A[A-Z]\w*/ post = $~.post_match if post.empty? true else constant_name? post end else false end end |
#current_frame ⇒ Object
502 503 504 |
# File 'lib/debug/thread_client.rb', line 502 def current_frame get_frame(@current_frame_index) end |
#deactivate ⇒ Object
126 127 128 |
# File 'lib/debug/thread_client.rb', line 126 def deactivate @step_tp.disable if @step_tp end |
#debug_cmd(cmds) ⇒ Object
1125 1126 1127 1128 1129 1130 1131 |
# File 'lib/debug/thread_client.rb', line 1125 def debug_cmd(cmds) DEBUGGER__.debug{ cmd, *args = *cmds args = args.map { |arg| DEBUGGER__.safe_inspect(arg) } "#{inspect} receives Cmd { type: #{cmd.inspect}, args: #{args} } from Session" } end |
#debug_event(ev, args) ⇒ Object
1112 1113 1114 1115 1116 1117 |
# File 'lib/debug/thread_client.rb', line 1112 def debug_event(ev, args) DEBUGGER__.debug{ args = args.map { |arg| DEBUGGER__.safe_inspect(arg) } "#{inspect} sends Event { type: #{ev.inspect}, args: #{args} } to Session" } end |
#debug_mode(old_mode, new_mode) ⇒ Object
1119 1120 1121 1122 1123 |
# File 'lib/debug/thread_client.rb', line 1119 def debug_mode(old_mode, new_mode) DEBUGGER__.debug{ "#{inspect} changes mode (#{old_mode} -> #{new_mode})" } end |
#debug_suspend(event) ⇒ Object
1133 1134 1135 1136 1137 |
# File 'lib/debug/thread_client.rb', line 1133 def debug_suspend(event) DEBUGGER__.debug{ "#{inspect} is suspended for #{event.inspect}" } end |
#default_frame_formatter(frame) ⇒ Object
66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 91 92 93 94 95 96 97 98 99 100 101 |
# File 'lib/debug/thread_client.rb', line 66 def default_frame_formatter frame call_identifier_str = case frame.frame_type when :block level, block_loc = frame.block_identifier args = frame.parameters_info if !args.empty? args_str = " {|#{assemble_arguments(args)}|}" end "#{colorize_blue("block")}#{args_str} in #{colorize_blue(block_loc + level)}" when :method ci = frame.method_identifier args = frame.parameters_info if !args.empty? args_str = "(#{assemble_arguments(args)})" end "#{colorize_blue(ci)}#{args_str}" when :c colorize_blue(frame.c_identifier) when :other colorize_blue(frame.other_identifier) end location_str = colorize(frame.location_str, [:GREEN]) result = "#{call_identifier_str} at #{location_str}" if return_str = frame.return_str result += " #=> #{colorize_magenta(return_str)}" end result end |
#evaluate_result(r) ⇒ Object
1016 1017 1018 1019 |
# File 'lib/debug/server_cdp.rb', line 1016 def evaluate_result r v = variable nil, r v[:value] end |
#event!(ev, *args) ⇒ Object
213 214 215 216 217 |
# File 'lib/debug/thread_client.rb', line 213 def event! ev, *args debug_event(ev, args) @q_evt << [self, @output, ev, generate_info, *args] @output = [] end |
#frame_eval(src, re_raise: false) ⇒ Object
414 415 416 417 418 419 420 421 422 423 424 425 426 427 428 429 430 431 432 433 434 435 436 437 438 439 440 |
# File 'lib/debug/thread_client.rb', line 414 def frame_eval src, re_raise: false @success_last_eval = false b = current_frame.eval_binding special_local_variables current_frame do |name, var| b.local_variable_set(name, var) if /\%/ !~ name end result = frame_eval_core(src, b) @success_last_eval = true result rescue SystemExit raise rescue Exception => e return yield(e) if block_given? puts "eval error: #{e}" e.backtrace_locations&.each do |loc| break if loc.path == __FILE__ puts " #{loc}" end raise if re_raise end |
#frame_eval_core(src, b) ⇒ Object
387 388 389 390 391 392 393 394 395 396 397 398 399 400 401 402 403 404 405 406 407 |
# File 'lib/debug/thread_client.rb', line 387 def frame_eval_core src, b saved_target_frames = @target_frames saved_current_frame_index = @current_frame_index if b f, _l = b.source_location tp_allow_reentry do b.eval(src, "(rdbg)/#{f}") end else frame_self = current_frame.self tp_allow_reentry do frame_self.instance_eval(src) end end ensure @target_frames = saved_target_frames @current_frame_index = saved_current_frame_index end |
#frame_str(i, frame: ) ⇒ Object
684 685 686 687 688 689 |
# File 'lib/debug/thread_client.rb', line 684 def frame_str(i, frame: @target_frames[i]) cur_str = (@current_frame_index == i ? '=>' : ' ') prefix = "#{cur_str}##{i}" frame_string = @frame_formatter.call(frame) "#{prefix}\t#{frame_string}" end |
#generate_info ⇒ Object
207 208 209 210 211 |
# File 'lib/debug/thread_client.rb', line 207 def generate_info return unless current_frame { location: current_frame.location_str, line: current_frame.location.lineno } end |
#get_frame(index) ⇒ Object
506 507 508 509 510 511 512 |
# File 'lib/debug/thread_client.rb', line 506 def get_frame(index) if @target_frames @target_frames[index] else nil end end |
#get_src(frame, max_lines:, start_line: nil, end_line: nil, dir: +1) ⇒ Object
442 443 444 445 446 447 448 449 450 451 452 453 454 455 456 457 458 459 460 461 462 463 464 465 466 467 468 469 470 471 472 473 474 475 476 477 478 479 480 481 482 483 |
# File 'lib/debug/thread_client.rb', line 442 def get_src(frame, max_lines:, start_line: nil, end_line: nil, dir: +1) if file_lines = frame.file_lines frame_line = frame.location.lineno - 1 lines = file_lines.map.with_index do |e, i| cur = i == frame_line ? '=>' : ' ' line = colorize_dim('%4d|' % (i+1)) "#{cur}#{line} #{e}" end unless start_line if frame.show_line if dir > 0 start_line = frame.show_line else end_line = frame.show_line - max_lines start_line = [end_line - max_lines, 0].max end else start_line = [frame_line - max_lines/2, 0].max end end unless end_line end_line = [start_line + max_lines, lines.size].min end if start_line != end_line && max_lines [start_line, end_line, lines] end else # no file lines nil end rescue Exception => e p e pp e.backtrace exit! end |
#inspect ⇒ Object
174 175 176 177 178 179 180 |
# File 'lib/debug/thread_client.rb', line 174 def inspect if bt = @thread.backtrace "#<DBG:TC #{self.id}:#{@mode}@#{bt[-1]}>" else # bt can be nil "#<DBG:TC #{self.id}:#{@mode}>" end end |
#internalProperty(name, obj) ⇒ Object
1021 1022 1023 1024 1025 1026 |
# File 'lib/debug/server_cdp.rb', line 1021 def internalProperty name, obj v = variable name, obj v.delete :configurable v.delete :enumerable v end |
#location ⇒ Object
56 57 58 |
# File 'lib/debug/thread_client.rb', line 56 def location current_frame&.location end |
#make_breakpoint(args) ⇒ Object
757 758 759 760 761 762 763 764 765 766 767 768 769 770 771 772 773 774 775 776 777 778 779 780 781 782 783 784 785 786 787 788 789 790 791 |
# File 'lib/debug/thread_client.rb', line 757 def make_breakpoint args case args.first when :method klass_name, op, method_name, cond, cmd, path = args[1..] bp = MethodBreakpoint.new(current_frame.eval_binding, klass_name, op, method_name, cond: cond, command: cmd, path: path) begin bp.enable rescue NameError => e if bp.klass puts "Unknown method name: \"#{e.name}\"" else # klass_name can not be evaluated if constant_name? klass_name puts "Unknown constant name: \"#{e.name}\"" else # only Class name is allowed puts "Not a constant name: \"#{klass_name}\"" bp = nil end end Session.activate_method_added_trackers if bp rescue Exception => e puts e.inspect bp = nil end bp when :watch ivar, object, result, cond, command, path = args[1..] WatchIVarBreakpoint.new(ivar, object, result, cond: cond, command: command, path: path) else raise "unknown breakpoint: #{args}" end end |
#management? ⇒ Boolean
130 131 132 |
# File 'lib/debug/thread_client.rb', line 130 def management? @is_management end |
#mark_as_management ⇒ Object
134 135 136 |
# File 'lib/debug/thread_client.rb', line 134 def mark_as_management @is_management = true end |
#name ⇒ Object
166 167 168 |
# File 'lib/debug/thread_client.rb', line 166 def name "##{@id} #{@thread.name || @thread.backtrace.last}" end |
#on_breakpoint(tp, bp) ⇒ Object
242 243 244 |
# File 'lib/debug/thread_client.rb', line 242 def on_breakpoint tp, bp suspend tp.event, tp, bp: bp end |
#on_init(name) ⇒ Object
234 235 236 |
# File 'lib/debug/thread_client.rb', line 234 def on_init name wait_reply [:init, name] end |
#on_load(iseq, eval_src) ⇒ Object
230 231 232 |
# File 'lib/debug/thread_client.rb', line 230 def on_load iseq, eval_src wait_reply [:load, iseq, eval_src] end |
#on_pause ⇒ Object
254 255 256 |
# File 'lib/debug/thread_client.rb', line 254 def on_pause suspend :pause end |
#on_trace(trace_id, msg) ⇒ Object
238 239 240 |
# File 'lib/debug/thread_client.rb', line 238 def on_trace trace_id, msg wait_reply [:trace, trace_id, msg] end |
#on_trap(sig) ⇒ Object
246 247 248 249 250 251 252 |
# File 'lib/debug/thread_client.rb', line 246 def on_trap sig if waiting? # raise Interrupt else suspend :trap, sig: sig end end |
#outline_method(o, klass, obj) ⇒ Object
714 715 716 717 718 719 720 721 722 723 724 725 726 |
# File 'lib/debug/thread_client.rb', line 714 def outline_method(o, klass, obj) begin singleton_class = M_SINGLETON_CLASS.bind_call(obj) rescue TypeError singleton_class = nil end maps = class_method_map((singleton_class || klass).ancestors) maps.each do |mod, methods| name = mod == singleton_class ? "#{klass}.methods" : "#{mod}#methods" o.dump(name, methods) end end |
#preview(name, obj) ⇒ Object
1092 1093 1094 1095 1096 1097 1098 1099 1100 1101 1102 1103 1104 1105 1106 1107 1108 1109 1110 1111 1112 1113 1114 |
# File 'lib/debug/server_cdp.rb', line 1092 def preview name, obj case obj when Array pd = propertyDescriptor name, obj overflow = false if obj.size > 100 obj = obj[0..99] overflow = true end hash = obj.each_with_index.to_h{|o, i| [i.to_s, o]} preview_ pd[:value], hash, overflow when Hash pd = propertyDescriptor name, obj overflow = false if obj.size > 100 obj = obj.to_a[0..99].to_h overflow = true end preview_ pd[:value], obj, overflow else nil end end |
#preview_(value, hash, overflow) ⇒ Object
1053 1054 1055 1056 1057 1058 1059 1060 1061 1062 1063 1064 1065 1066 1067 1068 |
# File 'lib/debug/server_cdp.rb', line 1053 def preview_ value, hash, overflow { type: value[:type], subtype: value[:subtype], description: value[:description], overflow: overflow, properties: hash.map{|k, v| pd = propertyDescriptor k, v { name: pd[:name], type: pd[:value][:type], value: pd[:value][:description] } } } end |
#process_cdp(args) ⇒ Object
775 776 777 778 779 780 781 782 783 784 785 786 787 788 789 790 791 792 793 794 795 796 797 798 799 800 801 802 803 804 805 806 807 808 809 810 811 812 813 814 815 816 817 818 819 820 821 822 823 824 825 826 827 828 829 830 831 832 833 834 835 836 837 838 839 840 841 842 843 844 845 846 847 848 849 850 851 852 853 854 855 856 857 858 859 860 861 862 863 864 865 866 867 868 869 870 871 872 873 874 875 876 877 878 879 880 881 882 883 884 885 886 887 888 889 890 891 892 893 894 895 896 897 898 899 900 901 902 903 904 905 906 907 908 909 910 911 912 913 914 915 916 917 918 919 920 921 922 923 924 925 926 927 928 929 930 931 932 933 934 935 936 937 938 939 940 941 942 943 944 945 946 947 948 949 950 951 952 953 954 955 956 957 958 959 960 961 962 963 964 965 966 967 968 969 970 971 972 973 974 975 976 977 978 979 980 981 982 983 984 985 986 987 988 989 990 991 992 993 994 995 996 997 |
# File 'lib/debug/server_cdp.rb', line 775 def process_cdp args type = args.shift req = args.shift case type when :backtrace exception = nil result = { reason: 'other', callFrames: @target_frames.map.with_index{|frame, i| exception = frame.raised_exception if frame == current_frame && frame.has_raised_exception path = frame.realpath || frame.path if frame.iseq.nil? lineno = 0 else lineno = frame.iseq.first_line - 1 end { callFrameId: SecureRandom.hex(16), functionName: frame.name, functionLocation: { # scriptId: N, # filled by SESSION lineNumber: lineno }, location: { # scriptId: N, # filled by SESSION lineNumber: frame.location.lineno - 1 # The line number is 0-based. }, url: path, scopeChain: [ { type: 'local', object: { type: 'object', objectId: rand.to_s } }, { type: 'script', object: { type: 'object', objectId: rand.to_s } }, { type: 'global', object: { type: 'object', objectId: rand.to_s } } ], this: { type: 'object' } } } } if exception result[:data] = evaluate_result exception result[:reason] = 'exception' end event! :cdp_result, :backtrace, req, result when :evaluate res = {} fid, expr, group = args frame = @target_frames[fid] = nil if frame && (b = frame.eval_binding) special_local_variables frame do |name, var| b.local_variable_set(name, var) if /\%/ !~name end result = nil case group when 'popover' case expr # Chrome doesn't read instance variables when /\A\$\S/ global_variables.each{|gvar| if gvar.to_s == expr result = eval(gvar.to_s) break false end } and ( = "Error: Not defined global variable: #{expr.inspect}") when /(\A((::[A-Z]|[A-Z])\w*)+)/ unless result = search_const(b, $1) = "Error: Not defined constant: #{expr.inspect}" end else begin result = b.local_variable_get(expr) rescue NameError # try to check method if b.receiver.respond_to? expr, include_all: true result = b.receiver.method(expr) else = "Error: Can not evaluate: #{expr.inspect}" end end end when 'console', 'watch-group' begin orig_stdout = $stdout $stdout = StringIO.new result = current_frame.binding.eval(expr.to_s, '(DEBUG CONSOLE)') rescue Exception => e result = e b = result.backtrace.map{|e| " #{e}\n"} frames = [ { columnNumber: 0, functionName: 'eval', lineNumber: 0, url: '' } ] e.backtrace_locations&.each do |loc| break if loc.path == __FILE__ path = loc.absolute_path || loc.path frames << { columnNumber: 0, functionName: loc.base_label, lineNumber: loc.lineno - 1, url: path } end res[:exceptionDetails] = { exceptionId: 1, text: 'Uncaught', lineNumber: 0, columnNumber: 0, exception: evaluate_result(result), stackTrace: { callFrames: frames } } ensure output = $stdout.string $stdout = orig_stdout end else = "Error: unknown objectGroup: #{group}" end else result = Exception.new("Error: Can not evaluate on this frame") end res[:result] = evaluate_result(result) event! :cdp_result, :evaluate, req, message: , response: res, output: output when :scope fid = args.shift frame = @target_frames[fid] if b = frame.binding vars = b.local_variables.map{|name| v = b.local_variable_get(name) variable(name, v) } special_local_variables frame do |name, val| vars.unshift variable(name, val) end vars.unshift variable('%self', b.receiver) elsif lvars = frame.local_variables vars = lvars.map{|var, val| variable(var, val) } else vars = [variable('%self', frame.self)] special_local_variables frame do |name, val| vars.unshift variable(name, val) end end event! :cdp_result, :scope, req, vars when :properties oid = args.shift result = [] prop = [] if obj = @obj_map[oid] case obj when Array result = obj.map.with_index{|o, i| variable i.to_s, o } when Hash result = obj.map{|k, v| variable(k, v) } when Struct result = obj.members.map{|m| variable(m, obj[m]) } when String prop = [ internalProperty('#length', obj.length), internalProperty('#encoding', obj.encoding) ] when Class, Module result = obj.instance_variables.map{|iv| variable(iv, obj.instance_variable_get(iv)) } prop = [internalProperty('%ancestors', obj.ancestors[1..])] when Range prop = [ internalProperty('#begin', obj.begin), internalProperty('#end', obj.end), ] end result += obj.instance_variables.map{|iv| variable(iv, obj.instance_variable_get(iv)) } prop += [internalProperty('#class', obj.class)] end event! :cdp_result, :properties, req, result: result, internalProperties: prop end end |
#process_dap(args) ⇒ Object
693 694 695 696 697 698 699 700 701 702 703 704 705 706 707 708 709 710 711 712 713 714 715 716 717 718 719 720 721 722 723 724 725 726 727 728 729 730 731 732 733 734 735 736 737 738 739 740 741 742 743 744 745 746 747 748 749 750 751 752 753 754 755 756 757 758 759 760 761 762 763 764 765 766 767 768 769 770 771 772 773 774 775 776 777 778 779 780 781 782 783 784 785 786 787 788 789 790 791 792 793 794 795 796 797 798 799 800 801 802 803 804 805 806 807 808 809 810 811 812 813 814 815 816 817 818 819 820 821 822 823 824 825 826 827 828 829 830 831 832 833 834 835 836 837 838 839 840 841 842 843 844 845 846 847 848 849 850 851 852 853 854 855 856 857 858 859 860 861 862 863 864 865 866 867 868 869 870 871 872 873 874 875 876 877 878 879 880 881 882 883 884 885 886 887 888 889 890 891 892 893 894 895 896 897 898 899 900 901 902 903 904 905 906 907 908 909 910 |
# File 'lib/debug/server_dap.rb', line 693 def process_dap args # pp tc: self, args: args type = args.shift req = args.shift case type when :backtrace start_frame = req.dig('arguments', 'startFrame') || 0 levels = req.dig('arguments', 'levels') || 1_000 frames = [] @target_frames.each_with_index do |frame, i| next if i < start_frame break if (levels -= 1) < 0 path = frame.realpath || frame.path source_name = path ? File.basename(path) : frame.location.to_s if (path && File.exist?(path)) && (local_path = UI_DAP.remote_to_local_path(path)) # ok else ref = frame.file_lines end frames << { id: i, # id is refilled by SESSION name: frame.name, line: frame.location.lineno, column: 1, source: { name: source_name, path: (local_path || path), sourceReference: ref, }, } end event! :dap_result, :backtrace, req, { stackFrames: frames, totalFrames: @target_frames.size, } when :scopes fid = args.shift frame = get_frame(fid) lnum = if frame.binding frame.binding.local_variables.size elsif vars = frame.local_variables vars.size else 0 end event! :dap_result, :scopes, req, scopes: [{ name: 'Local variables', presentationHint: 'locals', # variablesReference: N, # filled by SESSION namedVariables: lnum, indexedVariables: 0, expensive: false, }, { name: 'Global variables', presentationHint: 'globals', variablesReference: 1, # GLOBAL namedVariables: global_variables.size, indexedVariables: 0, expensive: false, }] when :scope fid = args.shift frame = get_frame(fid) vars = collect_locals(frame).map do |var, val| variable(var, val) end event! :dap_result, :scope, req, variables: vars, tid: self.id when :variable vid = args.shift obj = @var_map[vid] if obj case req.dig('arguments', 'filter') when 'indexed' start = req.dig('arguments', 'start') || 0 count = req.dig('arguments', 'count') || obj.size vars = (start ... (start + count)).map{|i| variable(i.to_s, obj[i]) } else vars = [] case obj when Hash vars = obj.map{|k, v| variable(value_inspect(k), v,) } when Struct vars = obj.members.map{|m| variable(m, obj[m]) } when String vars = [ variable('#length', obj.length), variable('#encoding', obj.encoding) ] when Class, Module vars = obj.instance_variables.map{|iv| variable(iv, obj.instance_variable_get(iv)) } vars.unshift variable('%ancestors', obj.ancestors[1..]) when Range vars = [ variable('#begin', obj.begin), variable('#end', obj.end), ] end vars += M_INSTANCE_VARIABLES.bind_call(obj).map{|iv| variable(iv, M_INSTANCE_VARIABLE_GET.bind_call(obj, iv)) } vars.unshift variable('#class', M_CLASS.bind_call(obj)) end end event! :dap_result, :variable, req, variables: (vars || []), tid: self.id when :evaluate fid, expr, context = args frame = get_frame(fid) = nil if frame && (b = frame.eval_binding) special_local_variables frame do |name, var| b.local_variable_set(name, var) if /\%/ !~ name end case context when 'repl', 'watch' begin result = b.eval(expr.to_s, '(DEBUG CONSOLE)') rescue Exception => e result = e end when 'hover' case expr when /\A\@\S/ begin result = M_INSTANCE_VARIABLE_GET.bind_call(b.receiver, expr) rescue NameError = "Error: Not defined instance variable: #{expr.inspect}" end when /\A\$\S/ global_variables.each{|gvar| if gvar.to_s == expr result = eval(gvar.to_s) break false end } and ( = "Error: Not defined global variable: #{expr.inspect}") when /\Aself$/ result = b.receiver when /(\A((::[A-Z]|[A-Z])\w*)+)/ unless result = search_const(b, $1) = "Error: Not defined constants: #{expr.inspect}" end else begin result = b.local_variable_get(expr) rescue NameError # try to check method if M_RESPOND_TO_P.bind_call(b.receiver, expr, include_all: true) result = M_METHOD.bind_call(b.receiver, expr) else = "Error: Can not evaluate: #{expr.inspect}" end end end else = "Error: unknown context: #{context}" end else result = 'Error: Can not evaluate on this frame' end event! :dap_result, :evaluate, req, message: , tid: self.id, **evaluate_result(result) when :completions fid, text = args frame = get_frame(fid) if (b = frame&.binding) && word = text&.split(/[\s\{]/)&.last words = IRB::InputCompletor::retrieve_completion_data(word, bind: b).compact end event! :dap_result, :completions, req, targets: (words || []).map{|phrase| detail = nil if /\b([_a-zA-Z]\w*[!\?]?)\z/ =~ phrase w = $1 else w = phrase end begin v = b.local_variable_get(w) detail ="(variable: #{value_inspect(v)})" rescue NameError end { label: phrase, text: w, detail: detail, } } else raise "Unknown req: #{args.inspect}" end end |
#propertyDescriptor(name, obj) ⇒ Object
1116 1117 1118 1119 1120 1121 1122 1123 1124 1125 1126 1127 1128 1129 1130 1131 1132 1133 1134 1135 1136 1137 1138 1139 1140 1141 1142 |
# File 'lib/debug/server_cdp.rb', line 1116 def propertyDescriptor name, obj case obj when Array propertyDescriptor_ name, obj, 'object', subtype: 'array' when Hash propertyDescriptor_ name, obj, 'object', subtype: 'map' when String propertyDescriptor_ name, obj, 'string', description: obj when TrueClass, FalseClass propertyDescriptor_ name, obj, 'boolean' when Symbol propertyDescriptor_ name, obj, 'symbol' when Integer, Float propertyDescriptor_ name, obj, 'number' when Exception bt = '' if log = obj.backtrace_locations log.each do |loc| break if loc.path == __FILE__ bt += " #{loc}\n" end end propertyDescriptor_ name, obj, 'object', description: "#{obj.inspect}\n#{bt}", subtype: 'error' else propertyDescriptor_ name, obj, 'object' end end |
#propertyDescriptor_(name, obj, type, description: nil, subtype: nil) ⇒ Object
1028 1029 1030 1031 1032 1033 1034 1035 1036 1037 1038 1039 1040 1041 1042 1043 1044 1045 1046 1047 1048 1049 1050 1051 |
# File 'lib/debug/server_cdp.rb', line 1028 def propertyDescriptor_ name, obj, type, description: nil, subtype: nil description = DEBUGGER__.safe_inspect(obj, short: true) if description.nil? oid = rand.to_s @obj_map[oid] = obj prop = { name: name, value: { type: type, description: description, value: obj, objectId: oid }, configurable: true, # TODO: Change these parts because enumerable: true # they are not necessarily `true`. } if type == 'object' v = prop[:value] v.delete :value v[:subtype] = subtype if subtype v[:className] = obj.class end prop end |
#puts(str = '') ⇒ Object
188 189 190 191 192 193 194 195 196 197 198 199 200 |
# File 'lib/debug/thread_client.rb', line 188 def puts str = '' if @recorder&. prefix = colorize_dim("[replay] ") end case str when nil @output << "\n" when Array str.each{|s| puts s} else @output << "#{prefix}#{str.chomp}\n" end end |
#puts_variable_info(label, obj, pat) ⇒ Object
598 599 600 601 602 603 604 605 606 607 608 609 610 611 612 613 614 615 616 617 618 619 620 621 |
# File 'lib/debug/thread_client.rb', line 598 def puts_variable_info label, obj, pat return if pat && pat !~ label begin inspected = DEBUGGER__.safe_inspect(obj) rescue Exception => e inspected = e.inspect end mono_info = "#{label} = #{inspected}" w = SESSION::width if mono_info.length >= w maximum_value_width = w - "#{label} = ".length valstr = truncate(inspected, width: maximum_value_width) else valstr = colored_inspect(obj, width: 2 ** 30) valstr = inspected if valstr.lines.size > 1 end info = "#{colorize_cyan(label)} = #{valstr}" puts info end |
#replay_suspend ⇒ Object
313 314 315 316 |
# File 'lib/debug/thread_client.rb', line 313 def replay_suspend # @recorder.current_position suspend :replay, replay_frames: @recorder.current_frame end |
#running? ⇒ Boolean
158 159 160 |
# File 'lib/debug/thread_client.rb', line 158 def running? @mode == :running end |
#search_const(b, expr) ⇒ Object
999 1000 1001 1002 1003 1004 1005 1006 1007 1008 1009 1010 1011 1012 1013 1014 |
# File 'lib/debug/server_cdp.rb', line 999 def search_const b, expr cs = expr.delete_prefix('::').split('::') [Object, *b.eval('Module.nesting')].reverse_each{|mod| if cs.all?{|c| if mod.const_defined?(c) mod = mod.const_get(c) else false end } # if-body return mod end } false end |
#set_mode(mode) ⇒ Object
138 139 140 141 142 143 144 145 146 147 148 149 150 151 152 153 154 155 156 |
# File 'lib/debug/thread_client.rb', line 138 def set_mode mode debug_mode(@mode, mode) # STDERR.puts "#{@mode} => #{mode} @ #{caller.inspect}" # pp caller # mode transition check case mode when :running raise "#{mode} is given, but #{mode}" unless self.waiting? when :waiting # TODO: there is waiting -> waiting # raise "#{mode} is given, but #{mode}" unless self.running? else raise "unknown mode: #{mode}" end # DEBUGGER__.warn "#{@mode} => #{mode} @ #{self.inspect}" @mode = mode end |
#show_by_editor(path = nil) ⇒ Object
cmd: show edit
633 634 635 636 637 638 639 640 641 642 643 644 645 646 647 648 649 650 651 652 653 |
# File 'lib/debug/thread_client.rb', line 633 def show_by_editor path = nil unless path if current_frame path = current_frame.path else return # can't get path end end if File.exist?(path) if editor = (ENV['RUBY_DEBUG_EDITOR'] || ENV['EDITOR']) puts "command: #{editor}" puts " path: #{path}" system(editor, path) else puts "can not find editor setting: ENV['RUBY_DEBUG_EDITOR'] or ENV['EDITOR']" end else puts "Can not find file: #{path}" end end |
#show_consts(pat, only_self: false) ⇒ Object
558 559 560 561 562 563 564 565 566 567 568 569 570 571 572 573 574 575 576 577 578 579 580 581 582 583 584 585 586 |
# File 'lib/debug/thread_client.rb', line 558 def show_consts pat, only_self: false if s = current_frame&.self cs = {} if M_KIND_OF_P.bind_call(s, Module) cs[s] = :self else s = M_CLASS.bind_call(s) cs[s] = :self unless only_self end unless only_self s.ancestors.each{|c| break if c == Object; cs[c] = :ancestors} if b = current_frame&.binding b.eval('Module.nesting').each{|c| cs[c] = :nesting unless cs.has_key? c} end end names = {} cs.each{|c, _| c.constants(false).sort.each{|name| next if names.has_key? name names[name] = nil value = c.const_get(name) puts_variable_info name, value, pat } } end end |
#show_frame(i = 0) ⇒ Object
680 681 682 |
# File 'lib/debug/thread_client.rb', line 680 def show_frame i=0 puts frame_str(i) end |
#show_frames(max = nil, pattern = nil) ⇒ Object
cmd: show frames
657 658 659 660 661 662 663 664 665 666 667 668 669 670 671 672 673 674 675 676 677 678 |
# File 'lib/debug/thread_client.rb', line 657 def show_frames max = nil, pattern = nil if @target_frames && (max ||= @target_frames.size) > 0 frames = [] @target_frames.each_with_index{|f, i| # we need to use FrameInfo#matchable_location because #location_str is for display # and it may change based on configs (e.g. use_short_path) next if pattern && !(f.name.match?(pattern) || f.matchable_location.match?(pattern)) # avoid using skip_path? because we still want to display internal frames next if skip_config_skip_path?(f.matchable_location) frames << [i, f] } size = frames.size max.times{|i| break unless frames[i] index, frame = frames[i] puts frame_str(index, frame: frame) } puts " # and #{size - max} frames (use `bt' command for all frames)" if max < size end end |
#show_globals(pat) ⇒ Object
589 590 591 592 593 594 595 596 |
# File 'lib/debug/thread_client.rb', line 589 def show_globals pat global_variables.sort.each{|name| next if SKIP_GLOBAL_LIST.include? name value = eval(name.to_s) puts_variable_info name, value, pat } end |
#show_ivars(pat) ⇒ Object
549 550 551 552 553 554 555 556 |
# File 'lib/debug/thread_client.rb', line 549 def show_ivars pat if s = current_frame&.self M_INSTANCE_VARIABLES.bind_call(s).sort.each{|iv| value = M_INSTANCE_VARIABLE_GET.bind_call(s, iv) puts_variable_info iv, value, pat } end end |
#show_locals(pat) ⇒ Object
543 544 545 546 547 |
# File 'lib/debug/thread_client.rb', line 543 def show_locals pat collect_locals(current_frame).each do |var, val| puts_variable_info(var, val, pat) end end |
#show_outline(expr) ⇒ Object
cmd: show outline
693 694 695 696 697 698 699 700 701 702 703 704 705 706 707 708 709 710 711 712 |
# File 'lib/debug/thread_client.rb', line 693 def show_outline expr begin obj = frame_eval(expr, re_raise: true) rescue Exception # ignore else o = Output.new(@output) locals = current_frame&.local_variables klass = M_CLASS.bind_call(obj) klass = obj if Class == klass || Module == klass o.dump("constants", obj.constants) if M_RESPOND_TO_P.bind_call(obj, :constants) outline_method(o, klass, obj) o.dump("instance variables", M_INSTANCE_VARIABLES.bind_call(obj)) o.dump("class variables", klass.class_variables) o.dump("locals", locals.keys) if locals end end |
#show_src(frame_index: @current_frame_index, update_line: false, max_lines: , **options) ⇒ Object
485 486 487 488 489 490 491 492 493 494 495 496 497 498 499 500 |
# File 'lib/debug/thread_client.rb', line 485 def show_src(frame_index: @current_frame_index, update_line: false, max_lines: CONFIG[:show_src_lines], **) if frame = get_frame(frame_index) start_line, end_line, lines = *get_src(frame, max_lines: max_lines, **) if start_line if update_line frame.show_line = end_line end puts "[#{start_line+1}, #{end_line}] in #{frame.pretty_path}" if !update_line && max_lines != 1 puts lines[start_line...end_line] else puts "# No sourcefile available for #{frame.path}" end end end |
#special_local_variables(frame) ⇒ Object
cmd: show
535 536 537 538 539 540 541 |
# File 'lib/debug/thread_client.rb', line 535 def special_local_variables frame SPECIAL_LOCAL_VARS.each do |mid, name| next unless frame&.send("has_#{mid}") name = name.sub('_', '%') if frame.eval_binding.local_variable_defined?(name) yield name, frame.send(mid) end end |
#step_tp(iter, events = [:line, :b_return, :return]) ⇒ Object
327 328 329 330 331 332 333 334 335 336 337 338 339 340 341 342 343 344 345 346 347 348 349 350 351 352 353 354 355 356 357 358 359 360 361 362 363 364 |
# File 'lib/debug/thread_client.rb', line 327 def step_tp iter, events = [:line, :b_return, :return] @step_tp.disable if @step_tp thread = Thread.current if SUPPORT_TARGET_THREAD @step_tp = TracePoint.new(*events){|tp| next if SESSION.break_at? tp.path, tp.lineno next if !yield(tp.event) next if tp.path.start_with?(__dir__) next if tp.path.start_with?('<internal:trace_point>') next unless File.exist?(tp.path) if CONFIG[:skip_nosrc] loc = caller_locations(1, 1).first next if skip_location?(loc) next if iter && (iter -= 1) > 0 tp.disable suspend tp.event, tp } @step_tp.enable(target_thread: thread) else @step_tp = TracePoint.new(*events){|tp| next if thread != Thread.current next if SESSION.break_at? tp.path, tp.lineno next if !yield(tp.event) next if tp.path.start_with?(__dir__) next if tp.path.start_with?('<internal:trace_point>') next unless File.exist?(tp.path) if CONFIG[:skip_nosrc] loc = caller_locations(1, 1).first next if skip_location?(loc) next if iter && (iter -= 1) > 0 tp.disable suspend tp.event, tp } @step_tp.enable end end |
#suspend(event, tp = nil, bp: nil, sig: nil, postmortem_frames: nil, replay_frames: nil, postmortem_exc: nil) ⇒ Object
258 259 260 261 262 263 264 265 266 267 268 269 270 271 272 273 274 275 276 277 278 279 280 281 282 283 284 285 286 287 288 289 290 291 292 293 294 295 296 297 298 299 300 301 302 303 304 305 306 307 308 309 310 311 |
# File 'lib/debug/thread_client.rb', line 258 def suspend event, tp = nil, bp: nil, sig: nil, postmortem_frames: nil, replay_frames: nil, postmortem_exc: nil return if management? debug_suspend(event) @current_frame_index = 0 case when postmortem_frames @target_frames = postmortem_frames @postmortem = true when replay_frames @target_frames = replay_frames else @target_frames = DEBUGGER__.capture_frames(__dir__) end cf = @target_frames.first if cf case event when :return, :b_return, :c_return cf.has_return_value = true cf.return_value = tp.return_value end if CatchBreakpoint === bp cf.has_raised_exception = true cf.raised_exception = bp.last_exc end if postmortem_exc cf.has_raised_exception = true cf.raised_exception = postmortem_exc end end if event != :pause show_src show_frames CONFIG[:show_frames] set_mode :waiting if bp event! :suspend, :breakpoint, bp.key elsif sig event! :suspend, :trap, sig else event! :suspend, event end else set_mode :waiting end wait_next_action end |
#to_s ⇒ Object
182 183 184 185 186 |
# File 'lib/debug/thread_client.rb', line 182 def to_s str = "(#{@thread.name || @thread.status})@#{current_frame&.location || @thread.to_s}" str += " (not under control)" unless self.waiting? str end |
#tp_allow_reentry ⇒ Object
369 370 371 372 373 374 375 376 377 378 379 380 |
# File 'lib/debug/thread_client.rb', line 369 def tp_allow_reentry TracePoint.allow_reentry do yield end rescue RuntimeError => e # on the postmortem mode, it is not stopped in TracePoint if e. == 'No need to allow reentrance.' yield else raise end end |
#truncate(string, width:) ⇒ Object
623 624 625 626 627 628 629 |
# File 'lib/debug/thread_client.rb', line 623 def truncate(string, width:) if string.start_with?("#<") string[0 .. (width-5)] + '...>' else string[0 .. (width-4)] + '...' end end |
#value_inspect(obj) ⇒ Object
688 689 690 691 |
# File 'lib/debug/server_dap.rb', line 688 def value_inspect obj # TODO: max length should be configuarable? DEBUGGER__.safe_inspect obj, short: true, max_length: 4 * 1024 end |
#variable(name, obj) ⇒ Object
1070 1071 1072 1073 1074 1075 1076 1077 1078 1079 1080 1081 1082 1083 1084 1085 1086 1087 1088 1089 1090 |
# File 'lib/debug/server_cdp.rb', line 1070 def variable name, obj pd = propertyDescriptor name, obj case obj when Array pd[:value][:preview] = preview name, obj obj.each_with_index{|item, idx| if valuePreview = preview(idx.to_s, item) pd[:value][:preview][:properties][idx][:valuePreview] = valuePreview end } when Hash pd[:value][:preview] = preview name, obj obj.each_with_index{|item, idx| key, val = item if valuePreview = preview(key, val) pd[:value][:preview][:properties][idx][:valuePreview] = valuePreview end } end pd end |
#variable_(name, obj, indexedVariables: 0, namedVariables: 0) ⇒ Object
937 938 939 940 941 942 943 944 945 946 947 948 949 950 951 952 953 954 |
# File 'lib/debug/server_dap.rb', line 937 def variable_ name, obj, indexedVariables: 0, namedVariables: 0 if indexedVariables > 0 || namedVariables > 0 vid = @var_map.size + 1 @var_map[vid] = obj else vid = 0 end ivnum = M_INSTANCE_VARIABLES.bind_call(obj).size { name: name, value: value_inspect(obj), type: (klass = M_CLASS.bind_call(obj)).name || klass.to_s, variablesReference: vid, indexedVariables: indexedVariables, namedVariables: namedVariables + ivnum, } end |
#wait_next_action ⇒ Object
796 797 798 799 800 |
# File 'lib/debug/thread_client.rb', line 796 def wait_next_action wait_next_action_ rescue SuspendReplay replay_suspend end |
#wait_next_action_ ⇒ Object
802 803 804 805 806 807 808 809 810 811 812 813 814 815 816 817 818 819 820 821 822 823 824 825 826 827 828 829 830 831 832 833 834 835 836 837 838 839 840 841 842 843 844 845 846 847 848 849 850 851 852 853 854 855 856 857 858 859 860 861 862 863 864 865 866 867 868 869 870 871 872 873 874 875 876 877 878 879 880 881 882 883 884 885 886 887 888 889 890 891 892 893 894 895 896 897 898 899 900 901 902 903 904 905 906 907 908 909 910 911 912 913 914 915 916 917 918 919 920 921 922 923 924 925 926 927 928 929 930 931 932 933 934 935 936 937 938 939 940 941 942 943 944 945 946 947 948 949 950 951 952 953 954 955 956 957 958 959 960 961 962 963 964 965 966 967 968 969 970 971 972 973 974 975 976 977 978 979 980 981 982 983 984 985 986 987 988 989 990 991 992 993 994 995 996 997 998 999 1000 1001 1002 1003 1004 1005 1006 1007 1008 1009 1010 1011 1012 1013 1014 1015 1016 1017 1018 1019 1020 1021 1022 1023 1024 1025 1026 1027 1028 1029 1030 1031 1032 1033 1034 1035 1036 1037 1038 1039 1040 1041 1042 1043 1044 1045 1046 1047 1048 1049 1050 1051 1052 1053 1054 1055 1056 1057 1058 1059 1060 1061 1062 1063 1064 1065 1066 1067 1068 1069 1070 1071 1072 1073 1074 1075 1076 1077 1078 1079 1080 1081 1082 1083 1084 1085 1086 1087 1088 1089 1090 1091 1092 1093 1094 1095 1096 1097 1098 1099 1100 1101 1102 1103 1104 1105 1106 1107 1108 1109 1110 |
# File 'lib/debug/thread_client.rb', line 802 def wait_next_action_ # assertions raise "@mode is #{@mode}" if !waiting? unless SESSION.active? pp caller set_mode :running return end while true begin set_mode :waiting if !waiting? cmds = @q_cmd.pop # pp [self, cmds: cmds] break unless cmds ensure set_mode :running end cmd, *args = *cmds case cmd when :continue break when :step step_type = args[0] iter = args[1] case step_type when :in if @recorder&. @recorder.step_forward raise SuspendReplay else step_tp iter do true end break end when :next frame = @target_frames.first path = frame.location.absolute_path || "!eval:#{frame.path}" line = frame.location.lineno if frame.iseq frame.iseq.traceable_lines_norec(lines = {}) next_line = lines.keys.bsearch{|e| e > line} if !next_line && (last_line = frame.iseq.last_line) > line next_line = last_line end end depth = @target_frames.first.frame_depth step_tp iter do loc = caller_locations(2, 1).first loc_path = loc.absolute_path || "!eval:#{loc.path}" # same stack depth (DEBUGGER__.frame_depth - 3 <= depth) || # different frame (next_line && loc_path == path && (loc_lineno = loc.lineno) > line && loc_lineno <= next_line) end break when :finish finish_frames = (iter || 1) - 1 goal_depth = @target_frames.first.frame_depth - finish_frames step_tp nil, [:return, :b_return] do DEBUGGER__.frame_depth - 3 <= goal_depth ? true : false end break when :back if @recorder&.can_step_back? unless @recorder.backup_frames @recorder.backup_frames = @target_frames end @recorder.step_back raise SuspendReplay else puts "Can not step back more." event! :result, nil end when :reset if @recorder&. @recorder.step_reset raise SuspendReplay end else raise "unknown: #{type}" end when :eval eval_type, eval_src = *args result_type = nil case eval_type when :p result = frame_eval(eval_src) puts "=> " + color_pp(result, 2 ** 30) if alloc_path = ObjectSpace.allocation_sourcefile(result) puts "allocated at #{alloc_path}:#{ObjectSpace.allocation_sourceline(result)}" end when :pp result = frame_eval(eval_src) puts color_pp(result, SESSION.width) if alloc_path = ObjectSpace.allocation_sourcefile(result) puts "allocated at #{alloc_path}:#{ObjectSpace.allocation_sourceline(result)}" end when :call result = frame_eval(eval_src) when :irb begin result = frame_eval('binding.irb') ensure # workaround: https://github.com/ruby/debug/issues/308 Reline.prompt_proc = nil if defined? Reline end when :display, :try_display failed_results = [] eval_src.each_with_index{|src, i| result = frame_eval(src){|e| failed_results << [i, e.] "<error: #{e.}>" } puts "#{i}: #{src} = #{result}" } result_type = eval_type result = failed_results else raise "unknown error option: #{args.inspect}" end event! :result, result_type, result when :frame type, arg = *args case type when :up if @current_frame_index + 1 < @target_frames.size @current_frame_index += 1 show_src max_lines: 1 show_frame(@current_frame_index) end when :down if @current_frame_index > 0 @current_frame_index -= 1 show_src max_lines: 1 show_frame(@current_frame_index) end when :set if arg index = arg.to_i if index >= 0 && index < @target_frames.size @current_frame_index = index else puts "out of frame index: #{index}" end end show_src max_lines: 1 show_frame(@current_frame_index) else raise "unsupported frame operation: #{arg.inspect}" end event! :result, nil when :show type = args.shift case type when :backtrace max_lines, pattern = *args show_frames max_lines, pattern when :list show_src(update_line: true, **(args.first || {})) when :edit show_by_editor(args.first) when :default pat = args.shift show_locals pat show_ivars pat show_consts pat, only_self: true when :locals pat = args.shift show_locals pat when :ivars pat = args.shift show_ivars pat when :consts pat = args.shift show_consts pat when :globals pat = args.shift show_globals pat when :outline show_outline args.first || 'self' else raise "unknown show param: " + [type, *args].inspect end event! :result, nil when :breakpoint case args[0] when :method bp = make_breakpoint args event! :result, :method_breakpoint, bp when :watch ivar, cond, command, path = args[1..] result = frame_eval(ivar) if @success_last_eval object = if b = current_frame.binding b.receiver else current_frame.self end bp = make_breakpoint [:watch, ivar, object, result, cond, command, path] event! :result, :watch_breakpoint, bp else event! :result, nil end end when :trace case args.shift when :object begin obj = frame_eval args.shift, re_raise: true opt = args.shift obj_inspect = DEBUGGER__.safe_inspect(obj) width = 50 if obj_inspect.length >= width obj_inspect = truncate(obj_inspect, width: width) end event! :result, :trace_pass, M_OBJECT_ID.bind_call(obj), obj_inspect, opt rescue => e puts e. event! :result, nil end else raise "unreachable" end when :record case args[0] when nil # ok when :on # enable recording if !@recorder @recorder = Recorder.new end @recorder.enable when :off if @recorder&.enabled? @recorder.disable end else raise "unknown: #{args.inspect}" end if @recorder&.enabled? puts "Recorder for #{Thread.current}: on (#{@recorder.log.size} records)" else puts "Recorder for #{Thread.current}: off" end event! :result, nil when :dap process_dap args when :cdp process_cdp args else raise [cmd, *args].inspect end end rescue SuspendReplay, SystemExit, Interrupt raise rescue Exception => e pp ["DEBUGGER Exception: #{__FILE__}:#{__LINE__}", e, e.backtrace] raise end |
#wait_reply(event_arg) ⇒ Object
events
221 222 223 224 225 226 227 228 |
# File 'lib/debug/thread_client.rb', line 221 def wait_reply event_arg return if management? set_mode :waiting event!(*event_arg) wait_next_action end |
#waiting? ⇒ Boolean
162 163 164 |
# File 'lib/debug/thread_client.rb', line 162 def waiting? @mode == :waiting end |